#include "ggen.h"
#include <math.h>

/*
 * routines that need implementation or major fixup:
 * circle_seg::limits(design *);
 * ellipse_seg::limits(design *);
 */

inline double square(double x) { return x*x; }

// =============== segment methods ==============

segment :: segment (void) { next=0; prev=0; status=0; }

void segment :: reverse (void) {
	cerr << "segment :: reverse : shouldn't be called\n";
}

void segment :: basic_reverse (void)
{
	// cout << "basic segment reversal" << endl;
	point pt;
	pt=start;
	start=end;
	end=pt;

	// angles
	double dt;
	dt=d_start;
	d_start=d_end+pi;
	d_end=dt+pi;

	segment *st;
	st=next;
	next=prev;
	prev=st;
}

void segment :: limits( design *d)
{
	cerr << "segment :: limits : shouldn't be called\n";
}

void segment :: set_status (int s)
{
	status = s;
}

line_seg :: line_seg (point _start, point _end) {
	start=_start;
	end=_end;
}

void line_seg :: reverse (void)
{
	this->basic_reverse();
}

void line_seg :: limits( design *d)
{
	d->push_limits(start);
	d->push_limits(end);
}

circle_seg :: circle_seg(point _start, point _end, point _cen, bool _dir) {
	start=_start;
	end=_end;
	center=_cen;
	dir=_dir;
	t_start = atan2(start.y-center.y, start.x-center.x);
	t_end   = atan2(  end.y-center.y,   end.x-center.x);
	radius= sqrt(square(start.y-center.y)+square(start.x-center.x));
}

// The alert reader will see that I stole these equations from xfig.3.2.2/u_geom.c
// The copyright license attached to that code requires that I display
// the following notice here:
/*
 * FIG : Facility for Interactive Generation of figures
 * Copyright (c) 1985-1988 by Supoj Sutanthavibul
 * Parts Copyright (c) 1989-1998 by Brian V. Smith
 * Parts Copyright (c) 1991 by Paul King
 *
 * Any party obtaining a copy of these files is granted, free of charge, a
 * full and unrestricted irrevocable, world-wide, paid up, royalty-free,
 * nonexclusive right and license to deal in this software and
 * documentation files (the "Software"), including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons who receive
 * copies from any such party to do so, with the only requirement being
 * that this copyright notice remain intact.
 *
 */
circle_seg :: circle_seg(point &p1, point &p2, point &p3)
{
	start = p1;
	end   = p3;

	if ((p1.x == p3.x && p1.y == p3.y) ||
	    (p1.x == p2.x && p1.y == p2.y) ||
	    (p2.x == p3.x && p2.y == p3.y)) {
		cout << "circle_seg::from three points failed test 1" << endl;
	}
	double  dx12 = p1.x - p2.x;
	double  dy12 = p1.y - p2.y;
	double  dx13 = p1.x - p3.x;
	double  dy13 = p1.y - p3.y;

	double s12 = asin(dy12 / sqrt(dx12 * dx12 + dy12 * dy12));
	double s13 = asin(dy13 / sqrt(dx13 * dx13 + dy13 * dy13));
	if (fabs(s12 - s13) < .001) {
		cout << "circle_seg::from three points failed test 2" << endl;
	}

	double len1 = p1.x * p1.x + p1.y * p1.y;
	double len2 = p2.x * p2.x + p2.y * p2.y;
	double len3 = p3.x * p3.x + p3.y * p3.y;
	double resy = (dx12 * (len3 - len1) - dx13 * (len2 - len1)) /
	   (2 * (dx13 * dy12 - dx12 * dy13));
	double resx;
	if (p1.x != p3.x)
		resx = (len3 + 2 * (resy) * dy13 - len1) / (2 * (-dx13));
	else
		resx = (len2 + 2 * (resy) * dy12 - len1) / (2 * (-dx12));

	center.x = resx;
	center.y = resy;
	// cout << "circle_seg::from three points determined center " << center << endl;
	radius = sqrt(square(p1.x-center.x)+square(p1.y-center.y));
	t_start = atan2(start.y-center.y, start.x-center.x);
	t_end   = atan2(  end.y-center.y,   end.x-center.x);
	dir     = 1;   // XXX fixme
}

void circle_seg :: reverse (void)
{
	this->basic_reverse();

	double at;
	at=t_start;
	t_start=t_end;
	t_end=at;
	dir=!dir;
}

void circle_seg :: limits( design *d)
{
	d->push_limits(start);
	d->push_limits(end);

	//                      positive y
	//                         q=1
	//        Quadrant II       |      Quadrant I
	//           i=1            |        i=0
	// q=2  -------------------------------------- q=0   positive x
	//        Quadrant III      |      Quadrant IV
	//           i=2            |        i=3
	//                         q=3

	int starti = (int) floor(t_start*2/pi);
	int endi   = (int) floor(t_end*2/pi);
	if (d->debug) cout << "circle_seg::limits " << starti << " " << endi << endl;
	for (int i = starti; i != endi; i = (i + (dir?1:-1))%4 ) {
		int q = (i + dir?1:0)%4;
		if (d->debug) cout << "  iteration " << i << " " << q << endl; 
		switch(q) {
			case 0:   d->push_limit_x(center.x+radius);  break;
			case 1:   d->push_limit_y(center.y+radius);  break;
			case 2:   d->push_limit_x(center.x-radius);  break;
			case 3:   d->push_limit_y(center.y-radius);  break;
		}
	}
}

ellipse_seg :: ellipse_seg(point _start, point _end, double _rx, double _ry) {
	start=_start;
	end=_end;
	rx=_rx;
	ry=_ry;
	circles=0;
}

void ellipse_seg :: reverse (void)
{
	this->basic_reverse();

	double at;
	at=t_start;
	t_start=t_end;
	t_end=at;
	dir=!dir;
	if (circles) cout << "Error: don't reverse after circle-izing ellipses" << endl;
}

void ellipse_seg :: limits( design *d)
{
	d->push_limits(start);
	d->push_limits(end);

	// XXX Need to copy code from circle_seg above
}

point *ellipse_seg :: ellipse_point (const double a)
{
	point *pp = new point(center.x+rx*cos(a), center.y+ry*sin(a) );
	return pp;
}

void ellipse_seg :: make_circles (void)
{
	if (circles) return;   // no fuss if it's already there
	int ncircles=3;  // for now
	double delta = (t_end - t_start)/ncircles;
	circles = new design();
	for (int i=0; i<ncircles; i++) {
		point *p1=ellipse_point(t_start + delta*(i));
		point *p3=ellipse_point(t_start + delta*(i+1));
#if 1
		point *p2=ellipse_point(t_start + delta*(i+0.5));
		circle_seg *c = new circle_seg(*p1, *p2, *p3);
		circles->add(c);
#else
		line_seg *l = new line_seg(*p1, *p3);
		circles->add(l);
#endif	
	}
}

// =============== ray methods ==============
ray::ray() {}

line_ray::line_ray(design *d, point _p, angle _a)
{
	p=_p;
	a=_a;
}

line_ray::line_ray(design *d, const ellipse_ray &_e, angle _a)
{
	// angle _a is the direction of the line, tangent to ellipse
	// angle f is the angle normal to ellipse 
	angle f = _a + ( _e.dir ? -M_PI/2 : M_PI/2);
	angle thp=atan2(_e.ry*sin(f),_e.rx*cos(f));
	point p2(_e.center.x+_e.rx*cos(thp),_e.center.y+_e.ry*sin(thp));
	if (d->debug) {
	cout << "line_ray from ellipse_ray and angle:" << endl;
	cout << "   input    a = " << _a << "   rx = " << _e.rx << "  ry = " << _e.ry << endl;
	cout << "   deduce   f = " << f  << "   thp = " << thp << "   p2 = " << p2 << endl;
	}
	ellipse_seg *e = new ellipse_seg(_e.p,p2,_e.rx,_e.ry);
	d->add(e);
	e->t_start = _e.ea;
	e->t_end  = thp;
	e->center = _e.center;
	e->dir    = _e.dir;
	p = p2;
	a = _a;
}

ellipse_ray::ellipse_ray(design *d, point _p, double _rx, double _ry, angle _a, bool _dir)
{
	p.x = _p.x;
	p.y = _p.y;
	rx  = _rx;
	ry  = _ry;
	a   = _a;
	double ba;
	if (_dir) ba = a-M_PI/2;
	     else ba = a+M_PI/2;
	ea  = atan2(ry*sin(ba),rx*cos(ba));
	center.x = p.x-_rx*cos(ea);
	center.y = p.y-_ry*sin(ea);
	if (d->debug) cout << "ellipse_ray: ba=" << ba << "  ea=" << ea << "  center=" << center << endl;
	dir = _dir;
}

void join(design *d, line_ray &a, line_ray &b)
{
	join(d,a,b,0.0);
}

void join(design *d, line_ray &a, line_ray &b, double r)
{
	if (d->debug) {
	cout << "join of line a " << a.p << ", theta = " << a.a << endl;
	cout << "    and line b " << b.p << ", theta = " << b.a << endl;
	cout << "    radius = " << r << endl;
	}

	segment *l;
	double xdiff = a.p.x - b.p.x;
	double ydiff = a.p.y - b.p.y;
	// matrix is [ ma mb; mc md ]
	double ma = -cos(a.a);
	double mb =  cos(b.a);
	double mc = -sin(a.a);
	double md =  sin(b.a);
	double det = ma*md - mb*mc;
	if (d->debug)
	cout << "det = " << det << " compared to sin(dtheta) = " << sin(a.a-b.a) << endl;
	if (fabs(det) < .001) {
		// could get circle's radius, but its position would be unstable
		l = new line_seg(a.p,b.p);
		l->set_status(1);  // this line is broken
		d->add(l);
		return;
	}

	// inverse is 1/det * [ md -mb; -mc ma ]
	// multiply inverse by [ xdiff; ydiff ]
	double s = ( md*xdiff - mb*ydiff)/det;
	double t = (-mc*xdiff + ma*ydiff)/det;
	if (d->debug) cout << "s=" << s << "   t=" << t << endl;
	if (s < 0 || t < 0) {
		l = new line_seg(a.p,b.p);
		l->set_status(1);  // this line is broken
		d->add(l);
		return;
	}
	if (r <= 0) {
		point isect( a.p.x - s*ma, a.p.y - s*mc);
		l = new line_seg(a.p,isect);
		d->add(l);
		l = new line_seg(isect,b.p);
		d->add(l);
		return;
	}
	double pfactor = fabs(tan(0.5*(a.a-b.a)));   // guaranteed > .001
	double pullback = r / pfactor;
	if (d->debug) cout << " pfactor=" << pfactor << "  pullback=" << pullback << endl;
	if (s-pullback < 0) {
		pullback=s;
		r = pullback * pfactor;
		cout << "% warning s" << endl;
	}
	if (t-pullback < 0) {
		pullback=t;
		r = pullback * pfactor;
		cout << "% warning t" << endl;
	}
	s = s-pullback;
	t = t-pullback;
	if (d->debug) cout << "s=" << s << "   t=" << t << endl;
	bool dir = det>0;
	double rr = dir ? r : -r ;
	point sp( a.p.x - s*ma, a.p.y - s*mc);
	point tp( b.p.x + t*mb, b.p.y + t*md);
	point cp( sp.x + rr*mc, sp.y - rr*ma);
	
	if (s) { l = new line_seg(a.p,sp);          d->add(l); }
	if (r) { l = new circle_seg(sp,tp,cp,dir);  d->add(l); }
	if (t) { l = new line_seg(tp,b.p);          d->add(l); }
}

// =============== design methods =============

design::design() {
	begin = 0;
	end = 0;
	debug = 0;
}

void design::add( segment *s )
{
	s->next = 0;
	s->prev = end;
	if (debug) {
	    if (end) {
		cout << "design::add connection should match:" << endl;
		cout << "  new:   " << s->start << endl;
		cout << "  end:   " << end->end << endl;
	    } else {
		cout << "design::add first segment:" << endl;
		cout << "  start: " << s->start << endl;
		cout << "  end:   " << s->end   << endl;
	    }
	}
	if (end) end->next = s;
	if (!begin) {xmin=xmax=s->start.x; ymin=ymax=s->start.y;}
	if (!begin) begin = s;
	end = s;
	s->limits(this);
}

void design::add( design *d )
{
	if (d->begin) d->begin->prev = end;
	if (end) end->next = d->begin;
	if (!begin && d->begin) {xmin=xmax=d->begin->start.x; ymin=ymax=d->begin->start.y;}
	if (!begin) begin = d->begin;
	end = d->end;
	if (d->begin) {
		if (d->xmax > xmax) xmax=d->xmax;
		if (d->xmin < xmin) xmin=d->xmin;
		if (d->ymax > ymax) ymax=d->ymax;
		if (d->ymin < ymin) ymin=d->ymin;
	}
}

void design::reverse() {
	segment *s, *next;
	for (s=begin; s; s=next) {
		next = s->next;
		// cout << "design::reverse " << s << " will become " << next << endl;
		s->reverse();
	}
	next=begin;
	begin=end;
	end=next;
}

void design::eliminate_ellipses( void )
{
	for (segment *s=begin; s; s=s->next) {
		ellipse_seg *e = dynamic_cast<ellipse_seg *>(s);
		if (e) e->make_circles();
	}
}

void design::push_limits(point &p)
{
	if (p.x < xmin) xmin=p.x;
	if (p.x > xmax) xmax=p.x;
	if (p.y < ymin) ymin=p.y;
	if (p.y > ymax) ymax=p.y;
}

void design::push_limit_x(double x)
{
	if (x < xmin) xmin = x;
	if (x > xmax) xmax = x;
}

void design::push_limit_y(double y)
{
	if (y < ymin) ymin = y;
	if (y > ymax) ymax = y;
}

// =============== transfer from segment to target =========

void segment     ::emit( target_t *t) { cerr << "whazzup" << endl; }
void design      ::emit( target_t *t) { t->emit(this); }
void line_seg    ::emit( target_t *t) { t->emit(this); }
void circle_seg  ::emit( target_t *t) { t->emit(this); }
void ellipse_seg ::emit( target_t *t) { t->emit(this); }

// ============== default target methods ==============

void target_t :: unhandled(const char *s) {
	// XXX eventually put back typeid(*this).name()
	cerr << "target (" << "):"
	     << " Unhandled " << s << endl;
}
void target_t :: emit(const ellipse_seg *e) {unhandled("ellipse");}
void target_t :: emit(const circle_seg  *c) {unhandled("circle");}
void target_t :: emit(const line_seg    *l) {unhandled("line");}
void target_t :: emit(const design      *d) {unhandled("design");}
