/*
 * to-do:
 *   privatize most class member variables
 */

#include <iostream.h>

typedef double coord;
typedef double angle;

class point {
    public:
	coord x;
	coord y;
	point(const coord _x, const coord _y) : x(_x), y(_y) {}
	point() : x(0), y(0) {}

};
inline ostream &operator << (ostream &os, const point &p)
{
	os << "( " << p.x << ", " << p.y << " )";
	return(os);
}
const double pi = 3.1415926535898;

class target_t;
class design;

// ============= basic segment and derived actual segments ==========

class segment {
    public:
	segment(void);
	virtual void emit( target_t *t);
	virtual void reverse( void );
	virtual void limits( design *d);
	void basic_reverse( void );
	void set_status( int );
	point start;
	point end;
	angle d_start;  // under-used
	angle d_end;    // under-used
	segment *next;
	segment *prev;
	int status;
};

class line_seg : public segment {
    public:
	line_seg(point _start, point _end);
	virtual void emit( target_t *t);
	virtual void reverse( void);
	virtual void limits( design *d);
};

class circle_seg : public segment {
    public:
	circle_seg(point _start, point _end, point _cen, bool _dir);
	circle_seg(point &_start, point &_mid, point &_end);
	virtual void emit( target_t *t);
	virtual void reverse( void);
	virtual void limits( design *d);
	point center;
	double radius;
	angle t_start;
	angle t_end;
	bool dir;   // true -> increasing theta
};

class ellipse_seg : public segment {
    public:
	ellipse_seg(void) {}
	ellipse_seg(point _start, point _end, double _rx, double _ry);
	void make_circles();
	virtual void emit( target_t *t);
	virtual void reverse( void);
	virtual void limits( design *d);
	point *ellipse_point (const double a);
	point center;
	double rx;
	double ry;
	// angles are in normalized space, as in:
	// p = center + ( rx*cos(t) , ry*sin(t) )
	angle t_start;
	angle t_end;
	design *circles;
	bool dir;   // true -> increasing theta
};

// ============== ray for generating segments ===========

class ray {
    public:
	ray();
	point p;
	angle a;
};

class line_ray;
class ellipse_ray;

class line_ray : public ray {
    public:
	line_ray(design *d, point _p, angle _a);
	line_ray(design *d, const ellipse_ray &_e, angle _a);
};

class ellipse_ray: public ray {
    public:
	ellipse_ray(design *d, point _p, double _rx, double _ry, angle _a, bool _dir);
	point center;
	double rx;
	double ry;
	angle ea;   // location of the point on the ellipse
	bool dir;   // true == clockwise
};

void join(design *d, line_ray &a, line_ray &b);
void join(design *d, line_ray &a, line_ray &b, double r);

// ============== design ===========

class design {
    public:
	segment *begin;
	segment *end;
	double xmin;
	double xmax;
	double ymin;
	double ymax;
	design();
	virtual void emit( target_t *t);
	void reverse();
	void add( segment *s );
	void add( design *d );
	void eliminate_ellipses( void );
	void push_limits(point &p);
	void push_limit_x( double x);
	void push_limit_y( double y);
	int debug;
};

class target_t {
    public:
	void unhandled(const char *s);
	virtual void emit(const ellipse_seg *e);
	virtual void emit(const circle_seg  *c);
	virtual void emit(const line_seg    *l);
	virtual void emit(const design      *d);
};
