#include <iostream.h>
#include <math.h>
#include "ggen.h"

// primary defects at the moment:
//    ellipses not handled nicely like circles
//    trailing comma at end of coordinate list

class g_pov : public target_t {
    public:
	g_pov(ostream &_os) { os = &_os; }
	virtual void emit(const design *d);
	virtual void emit(const ellipse_seg *e);
	virtual void emit(const circle_seg *c);
	virtual void emit(const line_seg *l);
	void emit(const segment *s);
	void bezier(point &p1, point &p2, point &p3, double fraction);
	void outyx(const double x, const double y);
	ostream *os;
	int num_segments;
	bool count_segments;
};

void g_pov::outyx(const double y, const double x)
{
	*os << "," << endl << "<" << y << "," << x << ">";
}

void g_pov::emit(const design *d)
{
	// cout << "g_pov design emission" << endl;
	segment *s=d->begin;
	if (!s) {
		cout << "what?  Null design?" << endl;
		return;
	}

	num_segments = 0;
	count_segments = true;
	for (; s; s=s->next) { s->emit(this); }
	count_segments = false;
	*os << "lathe { bezier_spline " << num_segments ;
	for (s=d->begin; s; s=s->next) { s->emit(this); }
	// s=d->end;
	// *os << "<" << s->end.y << "," << s->end.x << ">" << endl;
	*os << endl << "// use default finish " << endl;
	*os << "}" << endl;
}

void g_pov::emit(const segment *s)
{
}

void g_pov::emit(const line_seg *l)
{
	num_segments += 4;
	if (count_segments) return;
	double t;
	int i;
	for (i=0;i<4;i++) {
		// Map i/4.0 in (0,1) onto (0,1) emphasizing ends
# define MAP(x) (1.0-0.5*cos(M_PI*(x)))
		// t=MAP(i/4.0);
		t = i/3.0;
		outyx( l->start.y+(l->end.y-l->start.y)*t,
		       l->start.x+(l->end.x-l->start.x)*t );
	}
}

// Most of the Bezier math (but no code) was cribbed from gs4.03,
// in particular the routines arc_add and gx_path_add_partial_arc

void g_pov::emit(const circle_seg *c)
{
	double te = c->t_end;
	if ( c->dir && te < c->t_start) te += 2*M_PI;
	if (!c->dir && te > c->t_start) te -= 2*M_PI;
	/* number of bezier curves we use to approximate circle */
	int nb = (int) (1+fabs(c->t_start-te)/.7);
	num_segments += nb*4;
	if (count_segments) return;
	double t1, t2, t3;
	double dt = (te-c->t_start)/((double)nb);
	double fraction=(4.0/3.0) / (1 + sqrt(3-2*cos(dt)));
	double rscale=1.0/cos(0.5*dt);
	// *os << "dt = " << dt << "  fraction = " << fraction << "  rscale = " << rscale << endl;
	int i;
	for (i=0;i<nb;i++) {
		t1=c->t_start+(i+0.0)*dt;
		point p1(c->center.x+c->radius*cos(t1),
		         c->center.y+c->radius*sin(t1));
		t2=c->t_start+(i+0.5)*dt;
		point p2(c->center.x+c->radius*cos(t2)*rscale,
		         c->center.y+c->radius*sin(t2)*rscale);
		t3=c->t_start+(i+1.0)*dt;
		point p3(c->center.x+c->radius*cos(t3),
		         c->center.y+c->radius*sin(t3));
		bezier(p1,p2,p3,fraction);
	}
}

// Bezier math should match from circular case, with independent scaling
// of x and y

void g_pov::emit(const ellipse_seg *e)
{
	double te = e->t_end;
	if ( e->dir && te < e->t_start) te += 2*M_PI;
	if (!e->dir && te > e->t_start) te -= 2*M_PI;
	/* number of bezier curves we use to approximate circle */
	int nb = (int) (1+fabs(e->t_start-te)/.7);
	num_segments += nb*4;
	if (count_segments) return;
	double t1, t2, t3;
	double dt = (te-e->t_start)/((double)nb);
	double fraction=(4.0/3.0) / (1 + sqrt(3-2*cos(dt)));
	double rscale=1.0/cos(0.5*dt);
	// *os << "dt = " << dt << "  fraction = " << fraction << "  rscale = " << rscale << endl;
	int i;
	for (i=0;i<nb;i++) {
		t1=e->t_start+(i+0.0)*dt;
		point p1(e->center.x+e->rx*cos(t1),
		         e->center.y+e->ry*sin(t1));
		t2=e->t_start+(i+0.5)*dt;
		point p2(e->center.x+e->rx*cos(t2)*rscale,
		         e->center.y+e->ry*sin(t2)*rscale);
		t3=e->t_start+(i+1.0)*dt;
		point p3(e->center.x+e->rx*cos(t3),
		         e->center.y+e->ry*sin(t3));
		bezier(p1,p2,p3,fraction);
	}
}

void g_pov::bezier(point &p1, point &p2, point &p3, double fraction)
{
	// *os << "bezier: p1 = " << p1 << "  frac = " << fraction << endl;
	// *os << "        p2 = " << p2 << endl;
	// *os << "        p3 = " << p3 << endl;
	outyx( p1.y                       , p1.x                        );
	outyx( p1.y + fraction*(p2.y-p1.y), p1.x + fraction*(p2.x-p1.x) );
	outyx( p3.y + fraction*(p2.y-p3.y), p3.x + fraction*(p2.x-p3.x) );
	outyx( p3.y                       , p3.x                        );
}

target_t* get_pov(ostream &os)
{
	return new g_pov(os);
}
